# Generated by Django 2.1.9 on 2019-07-15 12:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0025_layer_settings")]

    operations = [
        migrations.CreateModel(
            name="LayerGroup",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("view", models.IntegerField()),
                ("label", models.CharField(max_length=255)),
                (
                    "parent",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="terra_layer.LayerGroup",
                    ),
                ),
            ],
        ),
        migrations.AlterUniqueTogether(
            name="layergroup", unique_together={("view", "label")}
        ),
    ]



# Generated by Django 2.0.13 on 2019-06-28 07:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0021_auto_20190627_1513")]

    operations = [
        migrations.AlterField(
            model_name="layer", name="view", field=models.IntegerField()
        )
    ]



from django_geosource.views import SourceModelViewset
from geostore.views import FeatureViewSet, LayerViewSet, LayerGroupViewsSet

from ..permissions import ReadOnly, SourcePermission


class GeoSourceModelViewset(SourceModelViewset):
    permission_classes = (SourcePermission,)


class GeostoreLayerViewSet(LayerViewSet):
    permission_classes = (ReadOnly,)


class GeostoreFeatureViewSet(FeatureViewSet):
    permission_classes = (ReadOnly,)


class GeostoreLayerGroupViewsSet(LayerGroupViewsSet):
    permission_classes = (ReadOnly,)



import factory

from ..models import Scene, Layer


class SceneFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Scene

    custom_icon = factory.django.ImageField()


class LayerFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Layer



# Generated by Django 2.0.13 on 2019-06-26 07:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0017_remove_layer_legend_enable")]

    operations = [
        migrations.RenameField(
            model_name="layer", old_name="legend_template", new_name="legends"
        )
    ]



from django.conf import settings

default_settings = getattr(settings, "TERRA_LAYER_STYLE_SETTINGS", {})

# Min height between circle legend labels, in pixels. Typically based on label police size.
DEFAULT_CIRCLE_MIN_LEGEND_HEIGHT = default_settings.get("circle_min_legend_height", 14)
DEFAULT_FILL_COLOR = default_settings.get("fill_color", "#0000cc")
DEFAULT_FILL_OPACITY = default_settings.get("fill_opacity", 0.4)
DEFAULT_STROKE_COLOR = default_settings.get("stroke_color", "#ffffff")
DEFAULT_STROKE_WIDTH = default_settings.get("stroke_width", 0.3)



# Generated by Django 2.0.13 on 2019-05-22 15:22

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [("django_geosource", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="FilterField",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("filter_type", models.IntegerField()),
                (
                    "filter_settings",
                    django.contrib.postgres.fields.jsonb.JSONField(default={}),
                ),
                (
                    "field",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_geosource.Field",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Layer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("view", models.IntegerField()),
                ("name", models.CharField(max_length=255)),
                ("order", models.IntegerField(default=0)),
                ("description", models.TextField(blank=True)),
                ("layer_style", models.TextField(blank=True)),
                ("legend_enable", models.BooleanField(default=False)),
                ("legend_template", models.TextField(blank=True)),
                ("table_enable", models.BooleanField(default=False)),
                ("table_export_enable", models.BooleanField(default=False)),
                ("popup_enable", models.BooleanField(default=False)),
                ("popup_template", models.TextField(blank=True)),
                ("minisheet_enable", models.BooleanField(default=False)),
                ("minisheet_template", models.TextField(blank=True)),
                ("filter_enable", models.BooleanField(default=False)),
                (
                    "filter_fields",
                    models.ManyToManyField(
                        through="terra_layer.FilterField", to="django_geosource.Field"
                    ),
                ),
                (
                    "source",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_geosource.Source",
                    ),
                ),
                (
                    "table_export_fields",
                    models.ManyToManyField(
                        related_name="in_layers_export", to="django_geosource.Field"
                    ),
                ),
                (
                    "table_fields",
                    models.ManyToManyField(
                        related_name="in_layers", to="django_geosource.Field"
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="filterfield",
            name="layer",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="terra_layer.Layer"
            ),
        ),
    ]



# Generated by Django 2.0.13 on 2019-06-25 12:36

import json

import django.contrib.postgres.fields.jsonb
from django.db import migrations


def set_json_defaults(apps, schema_editor):
    Layer = apps.get_model("terra_layer", "Layer")
    for layer in Layer.objects.all():
        try:
            json.loads(layer.legend_template)
        except json.JSONDecodeError:
            layer.legend_template = "[]"
            layer.save()


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0015_remove_filterfield_filter_type")]

    operations = [
        migrations.RunPython(set_json_defaults),
        migrations.AlterField(
            model_name="layer",
            name="legend_template",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=list),
        ),
    ]



from rest_framework import permissions


class LayerPermission(permissions.BasePermission):
    def has_permission(self, request, view):
        return not request.user.is_anonymous and request.user.has_terra_perm(
            "can_manage_layers"
        )


class ScenePermission(permissions.BasePermission):
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return True

        return not request.user.is_anonymous and request.user.has_terra_perm(
            "can_manage_layers"
        )


class SourcePermission(permissions.BasePermission):
    def has_permission(self, request, view):
        return not request.user.is_anonymous and request.user.has_terra_perm(
            "can_manage_sources"
        )


class ReadOnly(permissions.BasePermission):
    def has_permission(self, request, view):
        return request.method in permissions.SAFE_METHODS



# Generated by Django 2.1.9 on 2019-07-23 08:14

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0032_auto_20190718_1742")]

    operations = [
        migrations.AddField(
            model_name="layergroup",
            name="settings",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        )
    ]



# Generated by Django 2.0.13 on 2019-06-25 09:06

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0014_filterfield_label")]

    operations = [migrations.RemoveField(model_name="filterfield", name="filter_type")]



# Generated by Django 3.0.4 on 2020-03-16 11:24

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("terra_layer", "0048_auto_20200227_1342"),
    ]

    operations = [
        migrations.AddField(
            model_name="filterfield",
            name="display",
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name="filterfield",
            name="settings",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="scene",
            name="config",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]



# Generated by Django 2.0.13 on 2019-06-24 13:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0013_auto_20190620_1455")]

    operations = [
        migrations.AddField(
            model_name="filterfield",
            name="label",
            field=models.CharField(blank=True, max_length=255),
        )
    ]



import csv
from django.core.management.base import BaseCommand

from django_geosource.models import Source
from terra_layer.models import Layer


class Command(BaseCommand):
    help = "update a given layer with a given Source"

    def add_arguments(self, parser):
        parser.add_argument("layer", type=int, help="pk of the layer")
        parser.add_argument("source", action="store", help="name of the new source")
        parser.add_argument(
            "--matches",
            action="store",
            help="optionnal csv file, separated with a comma, with field matching between sources",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="dry run with actions outputed instead",
        )

    def handle(self, **options):
        layer_id = options.get("layer")
        source_name = options.get("source")
        matches = options.get("matches")
        dry_run = options.get("dry_run")

        if not Layer.objects.filter(id=layer_id).exists():
            self.stdout.write(self.style.ERROR(f"Layer {layer_id} does not exists"))
            return

        fields_matches = {}
        if matches:
            with open(matches, "r") as f:
                reader = csv.reader(f, delimiter=",")
                for row in reader:
                    old_name, new_name = row
                    fields_matches[old_name] = new_name

        layer = Layer.objects.get(id=layer_id)
        source = Source.objects.get(name=source_name)
        layer.replace_source(source, fields_matches, dry_run)



# Generated by Django 2.1.9 on 2019-07-23 11:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0033_layergroup_settings")]

    operations = [
        migrations.AddField(
            model_name="layer", name="in_tree", field=models.BooleanField(default=True)
        )
    ]



# Generated by Django 2.1.9 on 2019-07-18 14:01

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0030_layergroup_order")]

    operations = [
        migrations.AlterModelOptions(
            name="layergroup", options={"ordering": ["order"]}
        ),
        migrations.AlterField(
            model_name="layergroup",
            name="selectors",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=None, null=True
            ),
        ),
    ]



from hashlib import md5
import uuid

from django.core.cache import cache
from django.db import models, transaction
from django.contrib.postgres.fields import JSONField
from django.utils.functional import cached_property
from django.utils.text import slugify
from django_geosource.models import Source, Field
from rest_framework.reverse import reverse

from .utils import get_layer_group_cache_key
from .schema import JSONSchemaValidator, SCENE_LAYERTREE
from .style import generate_style_from_wizard


class Scene(models.Model):
    """ A scene is a group of data visualisation in terra-visu.
    It's also a main menu entry.
    """

    name = models.CharField(max_length=255, unique=True)
    slug = models.SlugField(max_length=255, unique=True)
    category = models.CharField(max_length=255, default="map")
    custom_icon = models.ImageField(
        max_length=255, upload_to="scene-icons", null=True, default=None
    )
    order = models.IntegerField(default=0, db_index=True)
    tree = JSONField(
        default=list, validators=[JSONSchemaValidator(limit_value=SCENE_LAYERTREE)]
    )
    config = JSONField(default=dict)

    def get_absolute_url(self):
        return reverse("scene-detail", args=[self.pk])

    def tree2models(self, current_node=None, parent=None, order=0):
        """
        Generate groups structure from admin layer tree.
        This is a recursive function to handle each step of process.

        :param current_node: current node of the tree
        :param parent: The parent group of current node
        :param order: Current order to keep initial json order
        :returns: Nothing
        """

        # Init case, we've just launch the process
        if current_node is None:
            current_node = self.tree
            self.layer_groups.all().delete()  # Clear all groups to generate brand new one

        if not parent:
            # Create a default unique parent group that is ignored at export
            parent = LayerGroup.objects.create(view=self, label="Root")

        if isinstance(current_node, list):
            for idx, child in enumerate(current_node):
                self.tree2models(current_node=child, parent=parent, order=idx)

        elif "group" in current_node:
            # Handle groups
            group = parent.children.create(
                view=self,
                label=current_node["label"],
                exclusive=current_node.get("exclusive", False),
                selectors=current_node.get("selectors"),
                settings=current_node.get("settings", {}),
                order=order,
            )

            if "children" in current_node:
                self.tree2models(current_node=current_node["children"], parent=group)

        elif "geolayer" in current_node:
            # Handle layers
            layer = Layer.objects.get(pk=current_node["geolayer"])
            layer.group = parent
            layer.order = order
            layer.save()

    def insert_in_tree(self, layer, parts, group_config=None):
        """ Add the layer in tree. Each parts are a group name to find inside the tree.
            Here we assume that missing groups are added at first position of current node
            We create missing group with default exclusive group configuration (should be corrected later if necessary)
        """

        group_config = group_config or {}

        current_node = self.tree
        last_group = None
        for part in parts:
            found = False
            # Search in groups
            for group in current_node:
                if group.get("group") and group["label"] == part:
                    last_group = group
                    current_node = group["children"]
                    found = True
                    break
            if not found:
                # Add the missing group part
                new_group = {"group": True, "label": part, "children": []}
                current_node.append(new_group)
                last_group = new_group
                current_node = new_group["children"]

        # Node if found (or created) we can add the geolayer now
        current_node.append({"geolayer": layer.id, "label": layer.name})

        if group_config and last_group:
            # And update tho config
            last_group.update(group_config)

        self.save()

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name)

        super().save(*args, **kwargs)
        self.tree2models()  # Generate LayerGroups according to the tree

    class Meta:
        ordering = ["order"]


class LayerGroup(models.Model):
    view = models.ForeignKey(
        Scene, on_delete=models.CASCADE, related_name="layer_groups"
    )
    label = models.CharField(max_length=255)
    parent = models.ForeignKey(
        "self", null=True, on_delete=models.CASCADE, related_name="children"
    )
    order = models.IntegerField(default=0)
    exclusive = models.BooleanField(default=False)
    selectors = JSONField(null=True, default=None)
    settings = JSONField(default=dict)

    class Meta:
        ordering = ["order"]


class Layer(models.Model):
    uuid = models.UUIDField(unique=True, default=uuid.uuid4, editable=False)
    source = models.ForeignKey(Source, on_delete=models.CASCADE, related_name="layers")

    group = models.ForeignKey(
        LayerGroup, on_delete=models.SET_NULL, null=True, related_name="layers"
    )
    name = models.CharField(max_length=255, blank=False)
    # Whether the layer is shown in tree or hidden
    in_tree = models.BooleanField(default=True)

    order = models.IntegerField(default=0)

    description = models.TextField(blank=True)

    layer_style = JSONField(default=dict)
    layer_style_wizard = JSONField(default=dict)
    settings = JSONField(default=dict)
    active_by_default = models.BooleanField(default=False)

    legends = JSONField(default=list)

    table_enable = models.BooleanField(default=False)
    table_export_enable = models.BooleanField(default=False)

    popup_enable = models.BooleanField(default=False)
    popup_template = models.TextField(blank=True)
    popup_minzoom = models.FloatField(default=0)
    popup_maxzoom = models.FloatField(default=22)

    minisheet_enable = models.BooleanField(default=False)
    minisheet_template = models.TextField(blank=True)

    highlight_color = models.CharField(max_length=255, blank=True)
    main_field = models.ForeignKey(
        Field, null=True, on_delete=models.CASCADE, related_name="is_main_of"
    )

    interactions = JSONField(default=list)

    fields = models.ManyToManyField(Field, through="FilterField")

    @property
    def style(self):
        return self.layer_style

    @cached_property
    def layer_identifier(self):
        return md5(f"{self.source.slug}-{self.pk}".encode("utf-8")).hexdigest()

    class Meta:
        ordering = ("order", "name")

    def save(self, **kwargs):
        if self.layer_style_wizard:
            style, legend_addition = generate_style_from_wizard(
                self, self.layer_style_wizard
            )
            self.layer_style = style
            if not self.legends:
                legend_addition["title"] = self.name
                self.legends = [legend_addition]
            else:
                self.legends[0].update(legend_addition)

        super().save(**kwargs)

        # Invalidate cache for layer group
        if self.group:
            cache.delete(get_layer_group_cache_key(self.group.view))

    def __str__(self):
        return f"Layer({self.id}) - {self.name}"

    @transaction.atomic()
    def replace_source(self, new_source, fields_matches=None, dry_run=False):
        fields_matches = fields_matches or {}
        # update old field if ones from the new source
        # remove it when not present in the new source
        for filter_field in self.fields_filters.all():
            # if not fields_matches provided or found, we check with the filter_field name
            field_name = fields_matches.get(
                filter_field.field.name, filter_field.field.name
            )
            if new_source.fields.filter(name=field_name).exists():
                new_field = new_source.fields.get(name=field_name)
                if dry_run:
                    print(f"{filter_field.field.name} replaced by {new_field.name}.")
                else:
                    filter_field.field = new_field
            else:
                if dry_run:
                    print(f"Old field {field_name} deleted.")
                else:
                    filter_field.delete()

        # fields in the new source that don't exist in the old one are created
        for field in new_source.fields.all():
            if (
                not self.fields_filters.filter(field__name=field.name).exists()
                and field.name not in fields_matches.values()
            ):
                if dry_run:
                    print(f"New FilterField {field.name} created.")
                else:
                    self.fields_filters.create(field=field)
        if dry_run:
            print(f"{self.source} replaced by {new_source}.")
        else:
            self.source = new_source
            self.save()


class CustomStyle(models.Model):
    layer = models.ForeignKey(
        Layer, on_delete=models.CASCADE, related_name="custom_styles"
    )
    source = models.ForeignKey(
        Source, on_delete=models.CASCADE, related_name="sublayers"
    )
    style = JSONField(default=dict)
    interactions = JSONField(default=list)

    @property
    def layer_identifier(self):
        return md5(
            f"{self.source.slug}-{self.source.pk}-{self.pk}".encode("utf-8")
        ).hexdigest()


class FilterField(models.Model):
    field = models.ForeignKey(Field, on_delete=models.CASCADE)
    layer = models.ForeignKey(
        Layer, on_delete=models.CASCADE, related_name="fields_filters"
    )
    label = models.CharField(max_length=255, blank=True)

    order = models.IntegerField(default=0)

    filter_enable = models.BooleanField(default=False)
    filter_settings = JSONField(default=dict)
    format_type = models.CharField(max_length=255, default=None, null=True)

    # Whether the field can be exported
    exportable = models.BooleanField(default=False)

    # Whether the field is available in the table
    shown = models.BooleanField(default=False)

    # Whether the field is displayed by default in table
    display = models.BooleanField(default=True)

    # Config for all non handled things
    settings = JSONField(default=dict)

    class Meta:
        ordering = ("order",)



# Generated by Django 2.1.9 on 2019-07-15 12:44

from django.db import migrations, models
import django.db.models.deletion


def forwards_func(apps, schema_editor):
    LayerGroup = apps.get_model("terra_layer", "LayerGroup")
    Layer = apps.get_model("terra_layer", "Layer")

    def get_or_create_group(view, groups):
        try:
            group_list, group_name = groups.rsplit("/", 1)
            parent = get_or_create_group(view, group_list)
        except ValueError:
            group_name = groups
            parent = None

        group, _ = LayerGroup.objects.get_or_create(
            view=view, label=group_name, defaults={"parent": parent}
        )
        return group

    for layer in Layer.objects.all():

        try:
            groups, name = layer.name.rsplit("/", 1)
        except ValueError:
            name = layer.name
            group = get_or_create_group(layer.view, " ")
        else:
            group = get_or_create_group(layer.view, groups)

        layer.name = name
        layer.group = group

        layer.save()


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0025_auto_20190715_1425")]

    operations = [
        migrations.AddField(
            model_name="layer",
            name="group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="terra_layer.LayerGroup",
            ),
        ),
        migrations.RunPython(forwards_func),
        migrations.AlterField(
            model_name="layer",
            name="group",
            field=models.ForeignKey(
                null=False,
                on_delete=django.db.models.deletion.CASCADE,
                to="terra_layer.LayerGroup",
            ),
            preserve_default=True,
        ),
    ]



# Generated by Django 2.0.13 on 2019-08-20 09:36

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0024_customstyle_interactions")]

    operations = [
        migrations.AddField(
            model_name="layer",
            name="settings",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        )
    ]



# Generated by Django 2.2.7 on 2019-12-12 10:29

from django.db import migrations, transaction


def layer2dict(layer):
    return {"geolayer": layer.id, "label": layer.name}


def group2dict(group):
    children = []

    for sub_group in group.children.all():
        children.append(group2dict(sub_group))

    for sub_layer in group.layers.all():
        children.append(layer2dict(sub_layer))

    return {
        "label": group.label,
        "group": True,
        "children": children,
        "exclusive": group.exclusive,
        "selectors": group.selectors,
        "settings": group.settings,
    }


@transaction.atomic
def group2scene_tree(apps, schema_editor):
    Scene = apps.get_model("terra_layer", "Scene")
    LayerGroup = apps.get_model("terra_layer", "LayerGroup")
    Layer = apps.get_model("terra_layer", "Layer")

    def tree2models(scene, current_node, parent=None, order=0):
        """
        Copied and adapted from the one in `models.py` to avoid further evolutions.
        """

        if not parent:
            # Create a default unique parent group that is ignored at export
            parent = LayerGroup.objects.create(view=scene, label="Root")

        if isinstance(current_node, list):
            for idx, child in enumerate(current_node):
                tree2models(scene, current_node=child, parent=parent, order=idx)

        elif "group" in current_node:
            # Handle groups
            group = parent.children.create(
                view=scene,
                label=current_node["label"],
                exclusive=current_node.get("exclusive", False),
                order=order,
            )

            if "children" in current_node:
                tree2models(scene, current_node=current_node["children"], parent=group)

        elif "geolayer" in current_node:
            # Handle layers
            layer = Layer.objects.get(pk=current_node["geolayer"])
            layer.group = parent
            layer.order = order
            layer.save()

    for scene in Scene.objects.all():
        tree = []
        for lg in LayerGroup.objects.filter(view=scene, parent=None):
            tree.append(group2dict(lg))
        # There is no sublayer at level 0 so no need to handle them
        scene.tree = tree

        scene.save()
        scene.layer_groups.all().delete()  # Clear all groups to generate brand new one
        tree2models(scene, tree)


class Migration(migrations.Migration):

    dependencies = [
        ("terra_layer", "0045_add_tree_2_scene"),
    ]

    operations = [
        migrations.RunPython(group2scene_tree),
    ]



# Generated by Django 2.0.13 on 2019-06-07 09:51

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0005_auto_20190606_1635")]

    operations = [
        migrations.AlterModelOptions(
            name="layer",
            options={"permissions": (("can_manage_layers", "Can manage layers"),)},
        ),
        migrations.AlterField(
            model_name="layer",
            name="layer_style",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]



# Generated by Django 2.0.13 on 2019-06-26 07:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0016_auto_20190625_1436")]

    operations = [migrations.RemoveField(model_name="layer", name="legend_enable")]



from django.db import transaction
from rest_framework import serializers
from rest_framework.fields import SerializerMethodField
from rest_framework.reverse import reverse
from rest_framework.serializers import ModelSerializer, PrimaryKeyRelatedField

from .models import CustomStyle, FilterField, Layer, Scene


class SceneListSerializer(ModelSerializer):
    url = serializers.CharField(source="get_absolute_url", read_only=True)
    layers_tree_url = SerializerMethodField()

    def get_layers_tree_url(self, obj):
        return reverse("layerview", args=[obj.slug])

    class Meta:
        model = Scene
        fields = (
            "id",
            "name",
            "slug",
            "category",
            "custom_icon",
            "url",
            "layers_tree_url",
            "order",
        )


class SceneDetailSerializer(ModelSerializer):
    slug = serializers.SlugField(required=False)
    icon = serializers.SerializerMethodField()

    def get_icon(self, obj):
        if obj.custom_icon:
            return obj.custom_icon.url

    class Meta:
        model = Scene
        fields = "__all__"


class FilterFieldSerializer(ModelSerializer):
    sourceFieldId = PrimaryKeyRelatedField(source="field", read_only=True)

    class Meta:
        model = FilterField
        exclude = ("layer",)


class CustomStyleSerializer(ModelSerializer):
    class Meta:
        model = CustomStyle
        exclude = ("layer",)


class LayerListSerializer(ModelSerializer):
    class Meta:
        model = Layer
        fields = ("id", "source", "group", "name", "order", "active_by_default")

    def to_representation(self, obj):
        return {
            **super().to_representation(obj),
            "view": obj.group.view.pk if obj.group else None,
        }


class LayerDetailSerializer(ModelSerializer):
    fields = FilterFieldSerializer(many=True, read_only=True, source="fields_filters")
    custom_styles = CustomStyleSerializer(many=True, read_only=True)
    group = PrimaryKeyRelatedField(read_only=True)

    @transaction.atomic
    def create(self, validated_data):
        instance = super().create(validated_data)

        # Update m2m through field
        self._update_nested(instance, "custom_styles", CustomStyleSerializer)
        self._update_m2m_through(instance, "fields", FilterFieldSerializer)

        return instance

    def to_representation(self, obj):
        return {
            **super().to_representation(obj),
            "view": obj.group.view.pk if obj.group else None,
        }

    @transaction.atomic
    def update(self, instance, validated_data):

        instance = super().update(instance, validated_data)

        # Update m2m through field
        self._update_m2m_through(instance, "fields", FilterFieldSerializer)
        self._update_nested(instance, "custom_styles", CustomStyleSerializer)

        return instance

    def _update_nested(self, instance, field, serializer):
        getattr(instance, field).all().delete()

        for value in self.initial_data.get(field, []):
            obj = serializer(data=value)
            if obj.is_valid(raise_exception=True):
                obj.save(layer=instance)

    def _update_m2m_through(self, instance, field, serializer):
        getattr(instance, field).clear()

        for index, value in enumerate(self.initial_data.get(field, [])):
            value["order"] = index  # Add order field

            obj = serializer(data=value)
            if obj.is_valid(raise_exception=True):
                obj.save(layer=instance)

    class Meta:
        model = Layer
        fields = "__all__"



import collections


def dict_merge(dct, merge_dct, add_keys=True):
    dct = dct.copy()
    if not add_keys:
        merge_dct = {k: merge_dct[k] for k in set(dct).intersection(set(merge_dct))}

    for k, _ in merge_dct.items():
        if (
            k in dct
            and isinstance(dct[k], dict)
            and isinstance(merge_dct[k], collections.Mapping)
        ):
            dct[k] = dict_merge(dct[k], merge_dct[k], add_keys=add_keys)
        else:
            dct[k] = merge_dct[k]

    return dct


def get_layer_group_cache_key(scene, extras=[]):
    """
    :param scene: The scene to be cached
    :return: The cache key
    :rtype: string
    """
    extras_joined = "-".join(extras)
    return f"terra-layer-{scene.pk}-{extras_joined}"



# Generated by Django 2.0.13 on 2019-06-18 13:00

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0011_auto_20190612_1721")]

    operations = [
        migrations.AlterField(
            model_name="layer",
            name="source",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="layers",
                to="django_geosource.Source",
            ),
        )
    ]



import inspect
import sys

from django.utils.functional import cached_property
from django_geosource.models import Source, WMTSSource
from rest_framework import serializers

DEFAULT_SOURCE_NAME = "terra"


class SourceSerializer(serializers.BaseSerializer):
    @classmethod
    def get_object_serializer(cls, obj):

        source = obj.source.get_real_instance()
        clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)

        for _, serializer in clsmembers:
            if (
                serializer.__module__ == __name__
                and serializer.Meta.model is source.__class__
            ):
                return serializer(obj)

        return cls(obj)

    @cached_property
    def source_object(self):
        return self.instance.source.get_real_instance()

    def to_representation(self, obj):
        return {
            **obj.style,
            "id": obj.layer_identifier,
            "source": DEFAULT_SOURCE_NAME,
            "source-layer": self.source_object.slug,
        }

    class Meta:
        model = Source


class WMTSSourceSerializer(SourceSerializer):
    def to_representation(self, obj):
        return {
            **obj.style,
            "id": obj.layer_identifier,
            "type": "raster",
            "minzoom": self.source_object.minzoom or 0,
            "maxzoom": self.source_object.maxzoom or 24,
            "source": {
                "type": "raster",
                "tileSize": self.source_object.tile_size,
                "tiles": [self.source_object.url],
            },
        }

    class Meta:
        model = WMTSSource



#!/usr/bin/env python

import os
from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))

README = open(os.path.join(HERE, "README.md")).read()
CHANGES = open(os.path.join(HERE, "CHANGES.md")).read()

tests_require = [
    "factory-boy",
    "flake8",
    "coverage",
]

setup(
    name="terra-layer",
    version=open(os.path.join(HERE, "terra_layer", "VERSION.md")).read().strip(),
    include_package_data=True,
    author="Makina Corpus",
    author_email="terralego-pypi@makina-corpus.com",
    description="Geographic data visualizer backend for terra-visu",
    long_description=README + "\n\n" + CHANGES,
    description_content_type="text/markdown",
    long_description_content_type="text/markdown",
    packages=find_packages(),
    url="https://github.com/Terralego/terra-layer.git",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=[
        "Django>=2.2",
        "djangorestframework",
        "django-geosource>=0.3",
        "django-geostore>=0.3.11",
        "django-terra-accounts>=0.3.6",
        "jsonschema",
    ],
    extras_require={"dev": tests_require},
)



# Generated by Django 2.1.9 on 2019-07-16 14:43

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("terra_layer", "0028_layergroup_exclusive")]

    operations = [
        migrations.AddField(
            model_name="layergroup",
            name="selectors",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=list),
        ),
        migrations.AlterField(
            model_name="layer",
            name="group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="layers",
                to="terra_layer.LayerGroup",
            ),
        ),
    ]



# Generated by Django 2.0.13 on 2019-09-12 14:52

from django.db import migrations, models
from django.core.exceptions import ObjectDoesNotExist
import django.db.models.deletion


def migrate_main_field(apps, schema_editor):
    Layer = apps.get_model("terra_layer", "Layer")
    for layer in Layer.objects.all():
        try:
            field_name = layer.settings.get("filters", {}).pop("mainField", None)
            layer.main_field = layer.source.fields.get(name=field_name)
        except ObjectDoesNotExist:
            layer.main_field = None
        layer.save()


class Migration(migrations.Migration):

    dependencies = [
        ("django_geosource", "0014_auto_20190821_1648"),
        ("terra_layer", "0038_layer_highlight_color"),
    ]

    operations = [
        migrations.AddField(
            model_name="layer",
            name="main_field",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="is_main_of",
                to="django_geosource.Field",
            ),
        ),
        migrations.RunPython(migrate_main_field),
    ]



# Generated by Django 2.1.9 on 2019-06-27 13:13

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("django_geosource", "0009_source_slug"),
        ("terra_layer", "0020_sublayer"),
    ]

    operations = [
        migrations.CreateModel(
            name="CustomStyle",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("style", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                (
                    "layer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="custom_styles",
                        to="terra_layer.Layer",
                    ),
                ),
                (
                    "source",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="sublayers",
                        to="django_geosource.Source",
                    ),
                ),
            ],
        ),
        migrations.RemoveField(model_name="sublayer", name="layer"),
        migrations.RemoveField(model_name="sublayer", name="source"),
        migrations.DeleteModel(name="SubLayer"),
    ]



from django.apps import AppConfig
from terra_accounts.permissions_mixins import PermissionRegistrationMixin


class TerraLayerConfig(PermissionRegistrationMixin, AppConfig):
    name = "terra_layer"

    permissions = (
        ("can_manage_layers", "Can manage layers"),
        ("can_manage_sources", "Can manage sources"),
    )
